package de.hwrBerlin.ameisenproblem.controller;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.math.BigDecimal;
import java.util.LinkedList;

import de.hwrBerlin.ameisenproblem.databeans.Coordinate;
import de.hwrBerlin.ameisenproblem.exceptionhandling.ShowableException;
import de.hwrBerlin.ameisenproblem.view.Constants;

/**
 * 
 * @author Jonathan Wiens, Benjamin Pfeiffer Diese Klasse bietet Funktionaliten fr das Einlesen von TSP (Traveling Salesman Problem)- Dateien.
 */
public class TSPFileParser {
    
    // Die gefundenen Header-Informationen als Liste
    private static LinkedList<String> headerInformation;
    
    

    /**
     * Diese Funktion liest eine Datei ein und bereitet die Informationen ber die Koordinaten entsprechend auf. Es kann unterschieden werden, ob es sich um ein Tour-File oder ein TSP-File handelt.
     * Das Tour File kann nur einegelesen werden wenn der TOUR_SECTION Tag benutzt wurde!
     * 
     * @param fileName
     *            Muss ein gltiger Filename sein.
     * @param isTour
     *            Gibt an (true) ob es sich um ein Tour-File handelt oder nicht (false)
     * @return Die Liste der Koordinaten wird als LinkedList<Coordinate> ausgegeben
     */
    
    public static LinkedList<Coordinate> parseFile(String fileName, boolean isTour) {
        if (fileName == null) {
            // Werfe eine Exception und benachrichtige den Benutzer
            ShowableException exception = new ShowableException();
            exception.setMessage("Bitte geben Sie einen gltigen Dateinamen an!");
            System.out.println("Filename was null");
        }
        
        // Variablen zum Einlesen
        String tempLine;
        BufferedReader br;
        
        // Lese die angegebene Datei
        try {
            br = new BufferedReader(new FileReader(fileName));
            
            // Versuche den Header zu interpretieren
            headerInformation = getHeader(br, isTour);
            
            /*
             * Der Header wurde bereits geparst, es msste jetzt der Koordinatenpart kommen.
             */

            LinkedList<Coordinate> tspPositions = new LinkedList<Coordinate>();
            
            /*
             * Wenn als letzte Zeile schon die Koordinatensektion (flschlicherweise) gelesen wurde dann muss die letzte Header Zeile unbedingt fr die Koordinatensektion bercksichtigt werden
             */

            if (headerInformation.getLast().contains(Constants.ALTERNATIVE_HEADER_END)) {
                /*
                 * Lsche Koordinatenzeile aus dem Header und Speichere die letze tempLine in der Koordinatenlise
                 */
                tspPositions.add(getCoordinateFromLine(headerInformation.removeLast(), isTour));
            }
            /*
             * Lese Datei weiter ein bis entweder der Tag "EOF" erscheint, oder die gesamte Datei zu Ende ist
             */

            /*
             * Wenn es sich um eine Tour handelt kann es passieren, dass die Koordinaten nebeneinander geschrieben werden, dieser Fall wird nun bercksichtigt!
             */
            if (isTour && !tourFileTypeIsNormal(fileName)) {
                while ((tempLine = br.readLine()) != null && (!tempLine.contains(Constants.FILE_END))) {
                    if (tempLine != null && !tempLine.trim().equals("")) {
                        tspPositions.addAll(getCoordinateFromSpecialTourLine(tempLine));
                    }
                }
                System.out.println("Spez");
            }
            // Es ist eine "normale" Tour
            else {
                while ((tempLine = br.readLine()) != null && (!tempLine.contains(Constants.FILE_END))) {
                    if (tempLine != null && !tempLine.trim().equals("")) {
                        tspPositions.add(getCoordinateFromLine(tempLine, isTour));
                    }
                }
                System.out.println("Normal");
                
            }
            
            /*
             * Wenn es sich um eine Tour handelt, lsche letzten Punkt da dieser laut Spezifikation doppelt ist, aber im Projekt nicht bentigt wird
             */
            if (isTour) {
                tspPositions.removeLast();
            }
            
            // Schliee BufferedReader
            br.close();
            
            return tspPositions;
            
        }
        catch (IOException e) {
            // Werfe eine Exception und benachrichtige den Benutzer
            ShowableException exception = new ShowableException();
            exception.setMessage("Fehler beim Einlesen der Koordinaten aus der Datei: " + fileName + " . Bitte das Handbuch lesen!");
            System.out.println("TSPFileParser" + ": Fehler beim Parsen der Datei: " + fileName);
            e.printStackTrace();
            throw exception;
        }
    }
    


    /**
     * Diese Methode liest einen Header ein.
     * 
     * @param br
     *            Es muss ein bereits initalisierter BufferedReader bergeben werden.
     * @param isTour
     *            Gibt an (true) ob es sich um ein Tour-File handelt oder nicht (false)
     * @return Die Informationen des Header werden als LinkedList zurckgegeben.
     */
    private static LinkedList<String> getHeader(BufferedReader br, boolean isTour) {
        if (br == null) {
            System.out.println("Es wurde ein leerer BufferedReader bergeben");
            return null;
        }
        
        // Versuche den Header zu interpretieren
        headerInformation = new LinkedList<String>();
        
        // Tag der (normalerweise) das Ende eines Ende der Header Section anzeigt. Default: TSP
        String standardHeaderEnd = Constants.TSP_HEADER_END;
        if (isTour) {
            standardHeaderEnd = Constants.TOUR_HEADER_END;
        }
        
        String tempLine;
        try {
            /*
             * Lese solange, bis Datei zuende, bis zum Tag (NODE_COORD_SECTION) oder alternativ "1 " kommt
             */
            // System.out.println("Header wird gelesen.");
            while ((tempLine = br.readLine()) != null && !tempLine.contains(standardHeaderEnd) && !tempLine.contains(Constants.ALTERNATIVE_HEADER_END)) {
                
                // Fge die aktuelle Zeile in die Header-Liste ein
                headerInformation.add(tempLine);
            }
        }
        catch (Exception e) {
            // Werfe eine Exception und benachrichtige den Benutzer
            ShowableException exception = new ShowableException();
            exception.setMessage("Fehler beim Einlesen der Koordinaten aus der Datei. Der Header kann nicht gelesen werden!");
            System.out.println("TSPFileParser" + ": Fehler beim Parsen des Headers!");
            e.printStackTrace();
            throw exception;
        }
        return headerInformation;
    }
    


    /**
     * Diese Methode verarbeitet eine TextZeile zu einer Koordinate. Die Koordinaten mssen folgendermaen vorliegen: Name XPos YPos
     * 
     * @param line
     *            Die zu verbeitende Textzeile wird bentigt.
     * @param isTour
     *            Gibt an (true) ob es sich um ein Tour-File handelt oder nicht (false)
     * @return Die Informationen aus der Textzeile verpackt in der Databean "Coordinate"
     */
    private static Coordinate getCoordinateFromLine(String line, boolean isTour) {
        // Zerteile den String in die Wertebereiche
        LinkedList<String> lineInformation = cutStringLine(line);
        
        // Erstelle Koordinate (temporr)
        Coordinate tempCoordinate = new Coordinate();
        // Fge den ersten Part als Namen an
        tempCoordinate.setName(lineInformation.get(0));
        // Fge den zweiten Part als X-Wert an, konvertiere vorher
        
        if (!isTour) {
            try {
                
                tempCoordinate.setxValue(new BigDecimal(lineInformation.get(1)).doubleValue());
                // Fge den ersten Part als Y-Wert an, konvertiere vorher
                tempCoordinate.setyValue(new BigDecimal(lineInformation.get(2)).doubleValue());
            }
            catch (NumberFormatException e) {
                // Werfe eine Exception und benachrichtige den Benutzer
                ShowableException exception = new ShowableException();
                exception.setMessage("Fehler beim Einlesen der Koordinaten aus der Datei. Bitte das Handbuch lesen!");
                System.out.println("Fehler bei Zeile:" + line + "XVal:" + lineInformation.get(2) + "YVal:" + lineInformation.get(3));
                throw exception;
            }
        }
        
        // Gebe die ausgelesene Koordinate zurck
        return tempCoordinate;
    }
    


    /**
     * Diese Methode zerteilt den bergebenen String anhand von den Leerstellen. Im Gegensatz zur Split-Methode werden alle Leerstellen(variabel) ignoriert.
     * 
     * @param line
     *            Die Line die zerteilt werden soll.
     * @return Alle gefundenen Teilstrings werden als Liste zurckgegeben.
     */
    public static LinkedList<String> cutStringLine(String line) {
        // Hier sollen alle Teilstrings gespeichert werden.
        LinkedList<String> strings = new LinkedList<String>();
        
        /*
         * Suche die bergebene Line ab, speichere nur in Liste wenn der Teilstring selbst nicht leer ist
         */

        for (int i = 0; i < line.trim().split(" ").length; i++) {
            if (!line.trim().split(" ")[i].equals("")) {
                strings.add(line.trim().split(" ")[i]);
            }
        }
        
        return strings;
    }
    


    /**
     * Diese Methode prft, ob eine Tour-Datei als eine "normale" Tour abgespeichert ist, oder die Koordinaten-Liste in der Datei in einer Zeile (nebeneinander) angordnet ist.
     * 
     * @param filename
     *            Der Dateiname der zu prfenden Datei
     * @return True, falls Koordinaten untereinander geschrieben sind, false wenn nebeneinander
     */
    public static boolean tourFileTypeIsNormal(String fileName) {
        
        // Variablen zum Einlesen
        // String tempLine;
        BufferedReader br;
        
        // Lese die angegebene Datei
        try {
            br = new BufferedReader(new FileReader(fileName));
            
            // Es handelt sich um eine Tour
            getHeader(br, true);
            
            /*
             * Der Header wurde bereits geparst, es msste jetzt der Koordinatenpart kommen. Es wird nun geprft, ob die Koordinaten nebeneinander liegen oder nicht.
             */

            // Punktenamen stehen nebeneinander
            if (cutStringLine(br.readLine()).size() >= 2) {
                br.close();
                return false;
            }
            
            else {
                br.close();
                return true;
            }
        }
        catch (Exception e) {
            System.out.println("TSP-FileParser" + "Fehler beim Tour-Prfer");
            return false;
        }
    }
    


    /**
     * Diese Methode verarbeitet eine TextZeile zu einer Koordinate. Die Koordinaten mssen folgendermaen vorliegen: Name XPos YPos
     * 
     * @param line
     *            Die zu verbeitende Textzeile wird bentigt.
     * @param isTour
     *            Gibt an (true) ob es sich um ein Tour-File handelt oder nicht (false)
     * @return Die Informationen aus der Textzeile verpackt in der Databean "Coordinate"
     */
    private static LinkedList<Coordinate> getCoordinateFromSpecialTourLine(String line) {
        // Zerteile den String in die Wertebereiche
        LinkedList<String> lineInformation = cutStringLine(line);
        
        // Liste von Koordinaten
        LinkedList<Coordinate> tempCoordinateList = new LinkedList<Coordinate>();
        
        // Gebe fr jede LineInformation eine Koordinate an und speichere diese in einer Liste
        for (String coordinateName : lineInformation) {
            Coordinate tempCoordinate = new Coordinate();
            tempCoordinate.setName(coordinateName);
            tempCoordinateList.add(tempCoordinate);
        }
        
        return tempCoordinateList;
        
    }
    


    // GETTER SETTER
    
    public static LinkedList<String> getHeaderInformation() {
        return headerInformation;
    }
    
    // /**
    // * Diese Funktion liest die HeaderInformation
    // * und sucht anhand des Tags "Dimension" die Anzahl der Koordinaten
    // * @return Die ganzahlige Anzahl der Koordinaten wird zurckgegeben.
    // * Wurde keine Information gefunden, wird "-1" zurckgegeben
    // */
    // public int searchCoordinateCountInHeader() {
    // if (getHeaderInformation() == null) {
    // // Error werfen
    // System.out.println("No Header Information Found");
    // }
    //
    // for (String headerLine : getHeaderInformation()) {
    // if (headerLine.contains(COORDINATE_COUNT_HEADER)) {
    // // Hole Int-Wert aus dem Header bzw. der Line und konvertiere zu
    // // int
    // return new BigInteger(headerLine.substring(11)).intValue();
    // }
    // }
    // // Meldung werfen keine Koordinatenlnge gefunden
    // //System.out.println("Cant find coordinate count in File " + fileName);
    // return -1;
    // }
    
}
